<?php

use Stripe\Stripe;


/**
 * Assemble and return the URL used for an authentication / connect-with-stripe link.
 * Includes some info about the user/seller as pre-populated.
 *
 * @return string
 */
function bt_stripe_connect_auth_url() {
    global $current_user;
	//error_log( __FUNCTION__ . ' -- current user = ' . print_r( $current_user, 1 ) );

    $connect_auth_url = '';

    // things to do if a shop manager user is signed in
    if ( bt_is_seller() ) {

        /**
         * Get seller info
         */
        $seller_post = bt_get_seller_post();
	    $seller_post_type_name = bt_get_post_type_name( $seller_post );
        //error_log( __FUNCTION__ . ' -- seller post = ' . print_r( $seller_post, 1 ) );
        $seller_business_name = $seller_post->post_title;
        if ( ! defined( 'BT_ENV') || BT_ENV != 'prod' ) {
            // non-production environment
            $seller_business_name .= ' TEST ' . date( 'Ymd-His' );
        }

        /**
         * build url for stripe authentication button
         */
        $connect_auth_url = 'https://connect.stripe.com/express/oauth/authorize';
        $connect_auth_url .= '?redirect_uri=' . get_the_permalink();
        $connect_auth_url .= '&client_id=' . STRIPE_CONNECT_CLIENT_ID;
        $connect_auth_url .= '&state=' . STRIPE_CONNECT_STATE;
        $connect_auth_url .= '&suggested_capabilities[]=transfers';
        //$connect_auth_url .= '&suggested_capabilities[]=card_payments';
        // pre-fill things we collect ahead of time (@see https://stripe.com/docs/connect/oauth-reference#prefill-fields)
        $connect_auth_url .= '&stripe_user[email]=' . $current_user->user_email;
	    if ( method_exists( '\App\Controllers\Partials\Content' . $seller_post_type_name, 'zip' ) ) {
		    $phone = call_user_func_array(
			    [
				    '\App\Controllers\Partials\Content' . $seller_post_type_name,
				    'phone',
			    ],
			    [
				    $seller_post,
			    ]
		    );
		    if ( ! empty( $phone ) ) {
			    $phone = preg_replace( '/[^0-9]/', '', $phone );
			    //error_log( __METHOD__ . ': called static method and got phone: ' . $phone );
			    if ( strlen( $phone ) >= 10 ) {
			    	// only the first 10 digits in case there are multiple numbers
			    	$phone = substr( $phone, 0, 10 );
				    $connect_auth_url .= '&stripe_user[phone_number]=' . $phone;
			    }
		    }
	    }
        $connect_auth_url .= '&stripe_user[first_name]=' . get_user_meta( $current_user->ID, 'first_name', true );
        $connect_auth_url .= '&stripe_user[last_name]=' . get_user_meta( $current_user->ID, 'last_name', true );
        $connect_auth_url .= '&stripe_user[business_name]=' . $seller_business_name;
        $connect_auth_url .= '&stripe_user[url]=' . get_the_permalink( $seller_post );
        $connect_auth_url .= '&stripe_user[country]=US';

    }

    return $connect_auth_url;
}


/**
 * Check for request parameters and take action based on them. Will:
 * - Confirm state matches, if state is part of request in auth process.
 * - If a code is present from auth process, it will:
 *   - bail if the user already has a stripe account id on file
 *   - else make the auth request to obtain stripe account id, etc. and store this in user meta.
 * - Perform a deauthorization if requested (remove access from stripe, clear user meta).
 * - Return some result html/message if there is something to report.
 *
 * @return string
 */
function bt_stripe_process_auth_request() {
    global $current_user;

    $result = '';

    // things to do if a shop manager user is signed in
    if ( bt_is_seller() ) {

        /**
         * actions to take when we load the page upon return from connection with stripe, where we have params:
         * - state returned for verification
         * - authorization_code for sending back to Stripe
         */
        if ( isset( $_REQUEST ) && ! empty( $_REQUEST ) ) {

            // check for matching state
            if ( isset( $_REQUEST['state'] ) && $_REQUEST['state'] != STRIPE_CONNECT_STATE ) {
                wp_die( 'Connect with Stripe: state is invalid.' );
            }

            // set our secret key for working with stripe
            \Stripe\Stripe::setApiKey( STRIPE_CONNECT_KEY_SECRET );

            // so long as we have state, check for code
            if ( isset( $_REQUEST['state'] ) && isset( $_REQUEST['code'] ) && ! empty( $_REQUEST['code'] ) ) {
                // we have a code for an authorization
                $code = $_REQUEST['code'];

                // make sure we don't already have authorization - visiting the success url would attempt again
                // and break the existing connection!
                if ( ! empty( bt_get_user_stripe_account_id() ) ) {
                    $result = '<div class="alert alert-warning">You are already connected to Stripe, so this attempt to connect has been stopped. You will see this message if you reload or revisit this URL after connecting.</div>';
                    return $result;
                }

                // post the code to stripe to get the connected account id
                //error_log( __FUNCTION__ . ' attempting to grant with code ' . $code );
                try {
                    $stripeResponse = \Stripe\OAuth::token( [
                        'grant_type' => 'authorization_code',
                        'code'       => $code,
                    ] );
                    //error_log( __FUNCTION__ . ' stripe response: ' . print_r( $stripeResponse, 1 ) );
                } catch ( \Stripe\Error\OAuth\InvalidGrant $e ) {
                    wp_die( 'Invalid authorization code: ' . $code );
                } catch ( \Exception $e ) {
                    error_log( __FUNCTION__ . ' exception: ' . print_r( $e->getError(), 1 ) );
                    wp_die( 'An unknown error occurred while authorizing the account. Check error log for details.' );
                }

                // store the connected account id to user meta
                if ( isset( $stripeResponse->stripe_user_id ) && ! empty( $stripeResponse->stripe_user_id ) ) {
                    update_user_meta( $current_user->ID, '_stripe_account_id', $stripeResponse->stripe_user_id );

                    $result = '<div class="alert alert-success">Connection with Stripe has been made successfully!</div>';

	                // approve seller if all criteria are now met
	                bt_approve_seller( $current_user->ID );
                }

            } else if ( isset( $_REQUEST['disconnect'] ) && ! empty( $_REQUEST['disconnect'] ) && $stripe_account_id = bt_get_user_stripe_account_id() ) {
                //error_log( __FUNCTION__ . ' attempting to revoke account ' . $stripe_account_id );
                try {
                    \Stripe\OAuth::deauthorize([
                        'client_id' => STRIPE_CONNECT_CLIENT_ID,
                        'stripe_user_id' => $stripe_account_id,
                    ]);
                } catch ( \Exception $e ) {
                    error_log( __FUNCTION__ . ' exception: ' . print_r( $e->getError(), 1 ) );
                    wp_die( 'An unknown error occurred while deauthorizing the account. Check error log for details.' );
                }

                // clear our custom meta
                bt_stripe_clear_user_meta();

                $result = '<div class="alert alert-success">Connection with Stripe has been deauthorized.</div>';
            }
        }
    }

    return $result;
}


/**
 * Retrieve the stripe account ID from user meta, if it exists.
 *
 * @param int|null $user_id A user ID to check. Will use current user if not provided.
 *
 * @return string
 */
function bt_get_user_stripe_account_id( $user_id = null ) {
    if ( empty( $user_id ) && ! bt_is_seller() ) {
        return '';
    } else if ( ! is_numeric( $user_id ) ) {
        global $current_user;
        $user_id = $current_user->ID;
    }

    return get_user_meta( $user_id, '_stripe_account_id', true );
}


/**
 * Deletes user meta associated with stripe.
 */
function bt_stripe_clear_user_meta() {
    global $current_user;

    delete_user_meta( $current_user->ID, '_stripe_account_id' );

    // un-approve seller
	$seller_post = bt_get_seller_post( $current_user->ID );
	update_field( 'seller_is_approved', '0', $seller_post );

	// send email alert to admin
	$user_name = bt_seller_contract_name( $current_user->ID );
	$user_obj = get_userdata( $current_user->ID );
	$user_email = $user_obj->user_email;

	// determine edit post and user URLs for admins to reference
	$edit_post_url = get_admin_url() . 'post.php?post=' . $seller_post->ID . '&action=edit';
	$edit_user_url = get_admin_url() . 'user-edit.php?user_id=' . $current_user->ID;

	// build email to alert about approval
	$subject = 'Seller Disconnected from Stripe on ' . get_bloginfo( 'name' );
	$message = <<<MSG
The following {$seller_post->post_type} has DISCONNECTED from Stripe, and thus has been un-approved.

Please review and make sure all is well.

Post: {$seller_post->post_title}
$edit_post_url

User: $user_name
$edit_user_url

Email: $user_email
MSG;

	$to = [
		'Curbside Marketplace <hello@curbsidemarketplace.com>',
	];
	$headers = [
		'From: ' . get_bloginfo( 'name' ) . ' Website <hello@curbsidemarketplace.com>',
	];

	// send the email
	if ( defined( 'BT_ENV') && BT_ENV == 'dev' ) {
		// just log if dev env
		error_log( 'SEND STRIPE DISCONNECTION EMAIL: ' . $subject . ' / ' . $message );
	} else {
		wp_mail( $to, $subject, $message, $headers );
		error_log( 'SENT STRIPE DISCONNECTION EMAIL: ' . $subject . ' / ' . $message );
	}

}


/**
 * Register a route for the REST API for processing payments to sellers.
 * An ID can be appended to the route end to get data for one specific seller.
 */
add_action( 'rest_api_init', function () {
    register_rest_route( 'bt/stripe/v1', '/seller-payouts(?:/(?P<id>\d+))?', array(
        'methods' => 'GET',
        'callback' => 'bt_stripe_seller_payouts',
        'args' => array(
            'id' => array(
                'default'=> 0,
                'validate_callback' => 'is_numeric',
                'sanitize_callback' => 'absint',
            ),
        ),
        'permission_callback' => function ( WP_REST_Request $request ) {
            return $request->get_header( 'x_request_signature' ) == REST_AUTH_TOKEN_BT_ECOMM;
        },
    ) );
} );


/**
 * Method to make payouts to sellers via REST request.
 *
 * @todo Convert this to use bt_payouts table and weekly processing scheme. This method was built before we
 * did weekly manual payouts, which calculates everything differently.
 *
 * @param WP_REST_Request $request
 *
 * @return stdClass|WP_Error
 * @throws Exception
 */
function bt_stripe_seller_payouts( WP_REST_Request $request ) {

    // set our secret key for working with stripe
    \Stripe\Stripe::setApiKey( STRIPE_CONNECT_KEY_SECRET );

    // init data for return
    $data = new stdClass();
    $data->sellers = [];

    // get all seller user accounts
    $wp_query_sellers = new \WP_User_Query( [
        'role__in' => [ 'shop_manager' ],
        'orderby' => 'ID',
        'order' => 'ASC',
        'number' => -1,
    ] );

    // process each seller
    foreach ( $wp_query_sellers->get_results() as $seller ) {
        $transfer_group = 'seller_' . $seller->ID . '_transfer_' . date( 'Ymd-His' );
        $seller_orders_total = 0;
        $platform_fee_total = 0;
        $platform_fee_rate = bt_seller_processing_fee_rate( $seller->ID );

        $seller_data = [
            'ID' => $seller->ID,
            'name' => trim( $seller->first_name . ' ' . $seller->last_name ),
            'transfer_group' => $transfer_group,
            'orders' => [],
        ];

        // get the seller's stripe account id from user meta
        $seller_stripe_account_id = bt_get_user_stripe_account_id( $seller->ID );
        // make sure we have stripe account id
        if ( empty( $seller_stripe_account_id ) ) {
            $message = 'Seller ID ' . $seller->ID . ' does NOT have a Stripe account ID!';
            \App\bt_email_alert( get_option( 'blogname' ) . ': Error processing seller payout', $message );
            continue;
        }

        // query for orders that do NOT have meta for payout, and which are more than X days old (for refund protection)
        $wp_query_orders = new \WP_Query( [
            'post_type' => 'shop_order',
            'author' => $seller->ID,
            'nopaging' => true,
            'posts_per_page' => 1,
            'post_status' => ['completed', 'wc-completed-picked', 'wc-completed-shipped' ],
            'orderby' => 'date',
            'order' => 'ASC',
            'date_query' => [
                'relation' => 'AND',
                'after' => '-14 days', // only those within the last couple of weeks, just in case.
                'before' => '-2 days', // get those that are more than two days ago, to help prevent refunds
            ],
            'meta_query' => [
                [
                    'relation' => 'OR',
                    [
                        'key'       =>  '_stripe_transfer_group',
                        'value'     =>  '',
                        'compare'   =>  '=',
                    ],
                    [
                        'key'       =>  '_stripe_transfer_group',
                        'compare'   =>  'NOT EXISTS',
                    ],
                ],
            ],
        ] );

        // calculate order fee, track data, and add up the total amount of money from the orders
        foreach ( $wp_query_orders->posts as $order ) {
            // make an order object and get the total
            $order_post = new \WC_Order( $order->ID );
            $order_total = (float) $order_post->get_total();

            // calculate platform fee for this order
            $platform_fee = round( $platform_fee_rate * $order_total, 2 );

            // add this order's total and platform fee to the seller's grand total
            $seller_orders_total += $order_total;
            $platform_fee_total += $platform_fee;

            // set dates
            $date_placed = $order_post->get_date_created();
            if ( is_object( $date_placed ) ) {
                $date_placed = $date_placed->date( 'Y-m-d H:i:s' );
            }
	        $date_completed = $order_post->get_date_completed();
	        if ( is_object( $date_completed ) ) {
		        $date_completed = $date_completed->date( 'Y-m-d H:i:s' );
	        }

            // add this order to the seller's data
            $seller_data['orders'][$order->ID] = [
                'ID' => $order->ID,
                'date_placed' => $date_placed,
                'date_completed' => $date_completed,
                'total' => $order_total,
                'platform_fee' => $platform_fee,
                //'WP_Post' => $order,
            ];
        }

        // calculate seller net total to be transferred
	    $seller_net = $seller_orders_total - $platform_fee_total;
	    $seller_net_no_decimal = round( $seller_net * 100 );

        // add totals for this transfer group to seller data
        $seller_data['order_count'] = count( $seller_data['orders'] );
        $seller_data['orders_grand_total'] = $seller_orders_total;
        $seller_data['orders_net_total'] = $seller_net;
        $seller_data['platform_fee_grand_total'] = $platform_fee_total;
        $seller_data['platform_fee_rate'] = $platform_fee_rate;

	    // init empty transfer result object, which is used if no transaction applies
	    $transfer = new \stdClass();
	    $transfer->id = '';

        // perform a transfer
	    if ( $seller_net_no_decimal > 0 ) {

		    // @todo check for available balance! and, how long does it take for funds to become available?

		    try {
			    $transfer = \Stripe\Transfer::create([
				    'amount' => $seller_net_no_decimal,
				    'currency' => 'usd',
				    'destination' => $seller_stripe_account_id,
				    // ONLY USE WHEN CHARGE TRANSACTIONS MADE WITH THIS, RIGHT? 'transfer_group' => $transfer_group,
			    ]);
		    } catch ( \Exception $e ) {
			    error_log( __FUNCTION__ . ' exception: ' . print_r( $e->getError(), 1 ) );
			    $message = 'An unknown error occurred while making a transfer for seller payout. Check error log for details.';
			    \App\bt_email_alert( get_option( 'blogname' ) . ': Error processing seller payout', $message );
			    wp_die( $message );
		    }
	    }

        // after successful transfer, loop orders again to record payout
        foreach ( $wp_query_orders->posts as $order ) {
            update_post_meta( $order->ID, '_stripe_transfer_group', $transfer_group );
            update_post_meta( $order->ID, '_stripe_platform_fee', $seller_data['orders'][$order->ID]['platform_fee'] );
            update_post_meta( $order->ID, '_stripe_platform_fee_rate', $platform_fee_rate );
        }

        // prep order data for database
	    $order_ids = '';
	    $orders = '';
	    if ( ! empty( $seller_data['orders'] ) ) {
		    $order_ids = serialize( array_keys( $seller_data['orders'] ) );
		    $orders = serialize( $seller_data['orders'] );
	    }

        // record payout data in our custom table
	    global $wpdb;
	    $wpdb->insert(
		    $wpdb->prefix . 'bt_stripe_payouts',
		    [
			    'user_id' => $seller->ID,
			    'transaction_id' => $transfer->id,
			    'time' => current_time( 'mysql' ),
			    'transfer_group' => $transfer_group,
			    'orders_total' => $seller_orders_total,
			    'orders_net' => $seller_net,
			    'platform_fee' => $platform_fee_total,
			    'platform_fee_rate' => $platform_fee_rate,
			    'order_count' => count( $seller_data['orders'] ),
			    'order_ids' => $order_ids,
			    'orders' => $orders,
		    ]
	    );

		// add the current seller data to the full data set
        $data->sellers[] = $seller_data;
    }

	//
    return $data;
}


/*
function bt_stripe_seller_payouts_report_records( $user = null ) {
	if ( empty( $user ) ) {
		$user = wp_get_current_user();
	}

	global $wpdb;

	// build sql to get payout records for this seller
	$sql = 'SELECT *
			FROM `' . $wpdb->prefix . 'bt_stripe_payouts`
			WHERE `user_id` = ' . $user->ID . '
			ORDER BY `time` DESC';

	$records = $wpdb->get_results( $sql );

	foreach ( $records as $key => $record ) {
		// unserialize order IDs
		$records[$key]->order_ids = unserialize( $record->order_ids );
		// unserialize orders
		$records[$key]->orders = unserialize( $record->orders );
	}

	return $records;
}
*/


/**
 * Filter to hide stripe fees and payouts for sellers' orders.
 *
 * @param bool $bool
 * @param int $order_id
 *
 * @return bool true if the item should be hidden
 */
function bt_stripe_hide_fees( $bool, $order_id ) {
	if ( bt_is_seller() ) {
		//error_log( 'STRIPE DISPLAY ORDER FEE - YES HIDE IT' );
		return true;
	} else {
		//error_log( 'STRIPE DISPLAY ORDER FEE - default' );
		return $bool;
	}
}
add_filter( 'wc_stripe_hide_display_order_fee', 'bt_stripe_hide_fees', 9988, 2 );
add_filter( 'wc_stripe_hide_display_order_payout', 'bt_stripe_hide_fees', 9988, 2 );


/**
 * Filter the payment intent request so that:
 * - We add a statement_descriptor_suffix if using sellers, which can be site-wide but defaults to seller name
 *
 * @param array $request
 * @param \WC_Order $order
 * @param object $prepared_source
 *
 * @return array The request data.
 */
function bt_filter_stripe_create_intent_request( $request, $order, $prepared_source ) {
	//error_log( __METHOD__ . ' -- request pre-altering = ' . print_r( $request, true ) );

	// things to do if using sellers
	if ( bt_use_sellers() ) {
		if ( $site_suffix = bt_use_sellers( 'stripe_statement_descriptor_suffix' ) ) {
			// use a site-wide suffix
			$request['statement_descriptor_suffix'] = $site_suffix;

		} else if ( ! empty( $seller = bt_get_order_seller_data( $order->get_id() ) ) ) {
			//error_log( __METHOD__ . ' -- it me. $seller = ' . print_r( $seller, true ) );
			// use seller name
			//error_log( 'STRIPE REQUEST seller id: ' . $seller->user_id . ' ' . $seller->post->post_title . ' / order id: ' . $order->get_id() );
			$request['statement_descriptor_suffix'] = substr( $seller->post->post_title, 0, 22 );
		}
	}

	//error_log( 'STRIPE REQUEST: ' . print_r( $request, 1 ) );
	return $request;
}
add_filter( 'wc_stripe_generate_create_intent_request', 'bt_filter_stripe_create_intent_request', 9988, 3 );
