<?php

/**
 * Things to do for when this plugin is for a site with Sellers.
 * - Add author support to Products and Orders
 * @todo Hook this in as appropriate so we can test for bt_use_sellers()
 */
add_post_type_support( 'product', 'author' );
add_post_type_support( 'wc_order', 'author' );
// THIS DIDN'T WORK AS EXPECTED: add_post_type_support( 'attachment', 'author' );


/**
 * If using sellers, require top-level admin capability to see Woo analytics.
 *
 * This disables the Woo breadcrumb/inbox/orders/etc. bar in wp-admin for sellers.
 */
add_filter( 'woocommerce_analytics_menu_capability', function( $capability ) {
  if ( bt_use_sellers() ) {
    $capability = 'level_9';
  }
  return $capability;
} );


/**
 * Check to see if the current site (theme via theme-supports) uses sellers, and optionally get a config item set in
 * the theme setup.
 *
 * @param string $config_item
 *
 * @return mixed false if no support, true if support but no requested config item, else the requested config item
 */
function bt_use_sellers( $config_item = '' ) {
  $seller_config = get_theme_support( 'bt-sellers' );
  //error_log( __METHOD__ . ' -- seller_config = ' . print_r( $seller_config, 1 ) );
  /* IS LIKE:
  seller_config = Array
  (
      [0] => Array
          (
              [help_email] => hello@curbsidemarketplace.com
              [email_from_use_marketplace_only] => 1
              [email_show_product_images] => 1
              ...
  */

  if ( empty( $seller_config ) ) {
    return false;
  }

  // get a config item if requested
  if ( ! empty( $config_item ) ) {
    foreach ( $seller_config as $config ) {
      //error_log( __METHOD__ . ' -- config = ' . print_r( $config, 1 ) );
      if ( array_key_exists( $config_item, $config ) ) {
        // found the config item in question
        //error_log( __METHOD__ . ' -- found = ' . $config[$config_item] );
        return $config[$config_item];
      }
    }
    // the specific config item requested wasn't found
    return false;
  }

  // theme support exists but no config requested
  return true;
}


/**
 * Check to see if there is a current user and if they have the shop_manager role, thus being a seller.
 *
 * @param null $user
 *
 * @return bool
 */
function bt_is_seller( $user = null ) {
  //error_log( __METHOD__ . ': user passed = ' . print_r( $user, 1 ) );
  if ( empty( $user ) ) {
    $user = wp_get_current_user();
    //error_log( __METHOD__ . ': user from get-current = ' . print_r( $user, 1 ) );
  }

  /*
  error_log( __METHOD__ . ' user = ' . $user->user_email
             . ' / is a \WP_User = ' . ( is_a( $user, '\WP_User' ) ? 'true' : 'false' )
             . ' / has shop_manager role = ' . ( in_array( 'shop_manager', (array) $user->roles ) ? 'true' : 'false' )
             . ' / bt_use_sellers = ' . ( bt_use_sellers() ? 'true' : 'false' )
    );
  */

  return
    is_a( $user, '\WP_User' )
    && in_array( 'shop_manager', (array) $user->roles )
    && bt_use_sellers();
}


/**
 * Check to see if there is a current user and if they have the subscriber role, thus potentially being a seller
 * if the site uses subscribers that could be sellers (e.g. cml/nma members which have seller-like profiles but no
 * e-commerce).
 *
 * @param null $user
 *
 * @return bool
 */
function bt_is_subscriber_could_be_seller( $user = null ) {
  //error_log( __METHOD__ . ': user passed = ' . print_r( $user, 1 ) );
  if ( empty( $user ) ) {
    $user = wp_get_current_user();
    //error_log( __METHOD__ . ': user from get-current = ' . print_r( $user, 1 ) );
  }

  /*
  error_log( __METHOD__ . ' user = ' . $user->user_email
             . ' / is a \WP_User = ' . ( is_a( $user, '\WP_User' ) ? 'true' : 'false' )
             . ' / has subscriber role = ' . ( in_array( 'subscriber', (array) $user->roles ) ? 'true' : 'false' )
             . ' / bt_use_sellers(subscribers_can_be_sellers) = ' . ( bt_use_sellers('subscribers_can_be_sellers') ? 'true' : 'false' )
    );
  */

  return
    is_a( $user, '\WP_User' )
    && in_array( 'subscriber', (array) $user->roles )
    && bt_use_sellers( 'subscribers_can_be_sellers' );
}


/**
 * Get and return the post that represents the seller, e.g. a brewery or artist.
 *
 * Will also check bt_is_subscriber_could_be_seller() so that we can get a profile post for a non-seller like in cml/nma.
 *
 * @param null $user_id
 *
 * @return \WP_Post|bool The Post object if it can be found, else false
 */
function bt_get_seller_post( $user_id = null ) {
  //error_log( __FUNCTION__ . ' -- running for user ID: ' . $user_id . ' / bt_is_seller() = ' . ( bt_is_seller() ? 'true' : 'false' ) . ' / bt_is_subscriber_could_be_seller() = ' . ( bt_is_subscriber_could_be_seller() ? 'true' : 'false' ) );

  // if no user id, meaning the current user, make sure they are a seller
  if ( ! $user_id && ! bt_is_seller() && ! bt_is_subscriber_could_be_seller() ) {
    //error_log( PHP_EOL . PHP_EOL . ' -- NO SELLER POST TO USE -- ' . PHP_EOL );
    return false;
  }

  if ( empty( $user_id ) ) {
    global $current_user;
    $user_id = $current_user->ID;
    //error_log( __FUNCTION__ . ' -- using CURRENT user! id = ' . $user_id );
  }

  $seller_post_type = bt_use_sellers( 'seller_post_type' );
  if ( ! empty( $seller_post_type ) ) {
    // query for the post recent post from the current user
    $args = [
      'post_type' => $seller_post_type, // could be array if multiple types exist
      'author' => $user_id,
      'nopaging' => true,
      'posts_per_page' => 1,
      'post_status' => ['publish', 'draft', 'pending'],
      'orderby' => 'date',
      'order' => 'DESC',
    ];
    $wp_query_seller = new \WP_Query( $args );

    if ( $wp_query_seller->have_posts() ) {
      return $wp_query_seller->posts[0];
    }
  }

  return false;
}


/**
 * Determine if seller does curbside pickup fulfillment. Default true, but will check controller method if exists.
 *
 * @param \WP_Post|null $seller
 *
 * @return bool
 */
function bt_seller_does_curbside( $seller = null ) {
  if ( ! $seller ) {
    $seller = bt_get_cart_seller();
    //error_log( __METHOD__ . ': arg was null, got CART seller ' . $seller->post_title );
  }
  if ( ! $seller ) {
    $seller = bt_get_seller_post();
    //error_log( __METHOD__ . ': arg was null, got seller POST ' . $seller->post_title );
  }

  $seller_post_type_name = bt_get_post_type_name( $seller );

  $does_curbside = true;

  if ( method_exists( '\App\Controllers\Partials\Content' . $seller_post_type_name, 'seller_does_curbside_static' ) ) {
    $does_curbside = call_user_func_array(
      [
        '\App\Controllers\Partials\Content' . $seller_post_type_name,
        'seller_does_curbside_static',
      ],
      [
        $seller,
      ]
    );
    //error_log( __METHOD__ . ': called static method and got: ' . $does_curbside );
  }

  //error_log( __METHOD__ . ': final result = ' . $does_curbside );
  return $does_curbside;
}


/**
 * Determine if seller does curbside by appointment only. Default false, but will check controller method if exists.
 *
 * @param \WP_Post|null $seller
 *
 * @return bool
 */
function bt_seller_does_curbside_by_appointment_only( $seller = null ) {
  if ( ! $seller ) {
    $seller = bt_get_cart_seller();
    //error_log( __METHOD__ . ': arg was null, got CART seller ' . $seller->post_title );
  }
  if ( ! $seller ) {
    $seller = bt_get_seller_post();
    //error_log( __METHOD__ . ': arg was null, got seller POST ' . $seller->post_title );
  }

  $seller_post_type_name = bt_get_post_type_name( $seller );

  $does_curbside_by_appointment_only = false;

  if ( method_exists( '\App\Controllers\Partials\Content' . $seller_post_type_name, 'seller_does_curbside_by_appointment_only_static' ) ) {
    $does_curbside_by_appointment_only = call_user_func_array(
      [
        '\App\Controllers\Partials\Content' . $seller_post_type_name,
        'seller_does_curbside_by_appointment_only_static',
      ],
      [
        $seller,
      ]
    );
    //error_log( __METHOD__ . ': called static method and got: ' . $does_curbside_by_appointment_only );
  }

  //error_log( __METHOD__ . ': final result = ' . $does_curbside_by_appointment_only );
  return $does_curbside_by_appointment_only;
}


/**
 * Determine if seller does USPS fulfillment. Default false, but will check controller method if exists.
 *
 * @param \WP_Post|null $seller
 *
 * @return bool
 */
function bt_seller_does_usps( $seller = null ) {
  if ( ! $seller ) {
    $seller = bt_get_cart_seller();
    //error_log( __METHOD__ . ': arg was null, got CART seller ' . $seller->post_title );
  }
  if ( ! $seller ) {
    $seller = bt_get_seller_post();
    //error_log( __METHOD__ . ': arg was null, got seller POST ' . $seller->post_title );
  }

  $seller_post_type_name = bt_get_post_type_name( $seller );

  $does_usps = false;

  if ( method_exists( '\App\Controllers\Partials\Content' . $seller_post_type_name, 'seller_does_usps_static' ) ) {
    $does_usps = call_user_func_array(
      [
        '\App\Controllers\Partials\Content' . $seller_post_type_name,
        'seller_does_usps_static',
      ],
      [
        $seller,
      ]
    );
    //error_log( __METHOD__ . ': called static method and got: ' . $does_usps );
  }

  //error_log( __METHOD__ . ': final result = ' . $does_usps );
  return $does_usps;
}


/**
 * Determine seller zip code
 *
 * @param \WP_Post|null $seller
 *
 * @return bool
 */
function bt_seller_zip( $seller = null ) {
  if ( ! $seller ) {
    $seller = bt_get_cart_seller();
    //error_log( __METHOD__ . ': arg was null, got CART seller ' . $seller->post_title );
  }
  if ( ! $seller ) {
    $seller = bt_get_seller_post();
    //error_log( __METHOD__ . ': arg was null, got seller POST ' . $seller->post_title );
  }

  $seller_post_type_name = bt_get_post_type_name( $seller );

  $zip = '';

  if ( method_exists( '\App\Controllers\Partials\Content' . $seller_post_type_name, 'zip' ) ) {
    $zip = call_user_func_array(
      [
        '\App\Controllers\Partials\Content' . $seller_post_type_name,
        'zip',
      ],
      [
        $seller,
      ]
    );
    //error_log( __METHOD__ . ': called static method and got: ' . $zip );
  }

  //error_log( __METHOD__ . ': final result = ' . $zip );
  return $zip;
}


/**
 * Find and return the post that corresponds to the seller in a cart (in-progress order). This is based on
 * cart session data, so it will build up session data if it isn't already set, by checking the author of
 * the first item in the cart.
 *
 * @return \WP_Post|null
 */
function bt_get_cart_seller() {
  if ( ! function_exists( 'WC' ) || empty( WC()->session ) || ! bt_use_sellers() ) {
    return null;
  }

  // see if we already have seller data in the cart session
  $seller_data = WC()->session->get( 'seller' );

  if ( empty( $seller_data ) ) {
    // we don't have a seller yet, so let's check the cart item(s) to see who they belong to
    $items = WC()->cart->get_cart_contents();

    if ( empty( $items ) ) {
      // no items, so no seller
      return null;
    }

    $first_item = current( $items );
    if ( empty( $first_item['product_id'] ) ) {
      // sanity check
      return null;
    }

    // author for the product
    $seller_user_id = bt_get_author_id_for_post( $first_item['product_id'] );

    //error_log( __METHOD__ . ' -- getting seller (id ' . $seller_user_id . ') data for item ' . $first_item['product_id'] );

    // seller post that corresponds to the author
    $seller = bt_get_seller_post( $seller_user_id );
    $seller_post_type_name = bt_get_post_type_name( $seller );

    // set seller data to store in cart session
    $seller_data['seller_id'] = $seller->ID; // post ID
    $seller_data['seller_user_id'] = $seller_user_id; // user ID
    $seller_data['seller_name'] = $seller->post_title;
    $seller_data['seller_logo'] = call_user_func_array(
      [
        '\App\Controllers\Partials\Content' . $seller_post_type_name,
        'logo',
      ],
      [
        'thumbnail',
        $seller->ID,
        true, // use site logo if none for seller
      ]
    );
    $seller_data['seller_url'] = get_the_permalink( $seller );

    // flag for curbside is available - default TRUE and check for a controller method to set seller-specific data
    $seller_data['seller_does_curbside'] = bt_seller_does_curbside( $seller );

    // flag for curbside is by appointment only - default FALSE and check for a controller method to set seller-specific data
    $seller_data['seller_does_curbside_by_appointment_only'] = bt_seller_does_curbside_by_appointment_only( $seller );

    // flag for usps is available - default FALSE and check for a controller method to set seller-specific data
    $seller_data['seller_does_usps'] = bt_seller_does_usps( $seller );

    /* this gets set fresh on each cart update action via bt_cart_update()
    if ( $location_slug = bt_use_sellers( 'seller_location_slug' ) ) {
      $locations = new \App\Locations( $seller->ID );
      $seller_data['seller_open_status'] = $locations->{$location_slug.'Status'}();
    }
    */

    // augment seller data with controller method from site/theme
    if ( method_exists( '\App\Controllers\Partials\Content' . $seller_post_type_name, 'custom_cart_seller_data' ) ) {
      $seller_data = call_user_func_array(
        [
          '\App\Controllers\Partials\Content' . $seller_post_type_name,
          'custom_cart_seller_data',
        ],
        [
          $seller_data,
        ]
      );
    }

    WC()->session->set( 'seller', $seller_data );
    // set a copy that will persist after emptying cart
    WC()->session->set( 'previous_seller', $seller_data );
    //error_log( __METHOD__ . ' -- saved fresh seller data to cart: ' . print_r( $seller_data, 1 ) );

  } else {
    // we have a seller in the cart session already; get the full seller post for the seller
    $seller = get_post( $seller_data['seller_id'] );
    //error_log( __METHOD__ . ' -- we already had seller data ' . print_r( $seller_data, 1 ) );
  }

  // return the post that corresponds to the seller
  return $seller;
}


/**
 * Determine the seller (user/author) ID for an order.
 *
 * If we need more than the ID, alternatives so far are:
 * - bt_get_order_seller_email()
 * - bt_get_order_seller_data() - has id, email, and post object
 *
 * @todo Save the seller ID to the cart session! Then just get it from there. Probably replaces bt_get_cart_seller() too.
 *
 * @param int $order_id
 *
 * @return string
 */
function bt_get_order_seller_id( $order_id ) {
  // create order and get the first product from it
  $order = wc_get_order( $order_id );
  $cart = $order->get_items();
  $cart_product = current( $cart );

  $seller_user_id = bt_get_author_id_for_post( $cart_product['product_id'] );

  //error_log( __METHOD__ . " product ID = $cart_product[product_id] / author-seller ID = $seller_user_id" );// / cart_product = " . print_r($cart_product,1) );

  // get and return the author id for the product
  return $seller_user_id;
}


/**
 * Get email address for user account that is a seller, given an order id.
 *
 * @param $order_id int
 *
 * @return string email address for the seller user account
 */
function bt_get_order_seller_email( $order_id ) {
  return get_the_author_meta( 'user_email', bt_get_order_seller_id( $order_id ) );
}


/**
 * Get various data for a seller, given an order id. For now we get:
 * - user_id: the WP user account ID
 * - user_email: the WP user account email address
 * - post: the post which corresponds to a seller
 *
 * @param $order_id
 *
 * @return \stdClass
 */
function bt_get_order_seller_data( $order_id ) {
  $data = new \stdClass();
  $data->user_id = bt_get_order_seller_id( $order_id );
  $data->user_email = sanitize_email( get_the_author_meta( 'user_email', $data->user_id  ) );
  $data->post = bt_get_seller_post( $data->user_id );
  return $data;
}


function bt_cart_seller_header( $order_id = null ) {
  if ( bt_use_sellers() ) {
    // get seller post
    if ( $order_id ) {
      //error_log( __METHOD__ . ' -- using order id: ' . $order_id );
      $seller_data = bt_get_order_seller_data( $order_id );
      $seller = $seller_data->post;
    } else {
      //error_log( __METHOD__ . ' -- using cart seller.' );
      $seller = bt_get_cart_seller();
    }
    //error_log( __METHOD__ . ' -- seller data = ' . print_r( $seller, 1 ) );

    if ( ! empty( $seller ) ) {
      $seller_post_type_name = bt_get_post_type_name( $seller );

      $use_default_logo = true;
      if ( bt_use_sellers( 'cart_header_use_default_logo' ) === false ) {
        // seller setup explicitly says don't use default logo
        $use_default_logo = false;
      }

      $seller_logo = call_user_func_array(
        [
          '\App\Controllers\Partials\Content' . $seller_post_type_name,
          'logo',
        ],
        [
          'thumbnail',
          $seller->ID,
          $use_default_logo, // use site logo if none for seller
        ]
      );

      // make seller meta using seller controller method specified in setup config, or default to city_state if there is none
      $seller_meta = '';
      $cart_header_meta_method = bt_use_sellers( 'cart_header_meta_method' );
      if ( empty( $cart_header_meta_method ) ) {
        // default to location
        $cart_header_meta_method = 'city_state';
      }
      if ( ! empty( $cart_header_meta_method ) && method_exists( '\App\Controllers\Partials\Content' . $seller_post_type_name, $cart_header_meta_method ) ) {
        $seller_meta = call_user_func_array(
          [
            '\App\Controllers\Partials\Content' . $seller_post_type_name,
            $cart_header_meta_method,
          ],
          [
            $seller->ID,
          ]
        );
      }
      if ( ! bt_use_sellers( 'omit_cart_seller_header_logo_and_title' ) ) {
        ?>
        <div class="cart-seller d-flex align-items-center justify-content-start mb-4">
          <?php if ( ! empty( $seller_logo ) ) : ?>
            <div class="logo mr-3">
              <?= $seller_logo ?>
            </div>
          <?php endif; ?>
          <div class="text">
            <p class="h5 mb-2">
              <?= $seller->post_title ?>
            </p>
            <?php if ( ! empty( $seller_meta ) ) : ?>
              <p class="meta mb-0">
                <?= $seller_meta ?>
              </p>
            <?php endif; ?>
          </div>
        </div>
        <?php
      }

      if ( is_cart() && bt_seller_does_curbside() ) {
        // get next pickup info via locations (we get this fresh all the time so it's current to the last cart action)
        if ( bt_seller_does_curbside_by_appointment_only( $seller ) ) {
          echo '
		                <div class="same-day-cutoff">
		                	<p>Curbside pickups are by appointment. Please indicate desired pickup times when you check out.</p>
		                </div>
		            ';
        } else if ( $location_slug = bt_use_sellers( 'seller_location_slug' ) ) {
          $locations = new \App\Locations( $seller->ID );
          $pick_up_message = $locations->{$location_slug.'SameDayCutoff'}();
          echo '
		                <div class="same-day-cutoff">
		                	' . $pick_up_message . '
		                </div>
		            ';
        }
      }
    }
  }
}


/**
 * Determine the name used for the seller contract. Will use "firstname lastname" but if that's empty it will use
 * nickname.
 *
 * @param null $user_id
 * @param bool $sanitized Run the name through sanitize_title() for comparison to name entered on contract.
 *
 * @return bool|string
 */
function bt_seller_contract_name( $user_id = null, $sanitized = false ) {
  if ( empty( $user_id ) ) {
    $user_id = get_current_user_id();
  }

  if ( empty( $user_id ) ) {
    return false;
  }

  // try the explicit custom field first, which we use when someone needs to use something different from
  // their user account name
  $contract_name = get_user_meta( $user_id, 'contract_name', true );
  if ( empty( $contract_name ) ) {
    // there isn't an explicit contract name, so construct with first/last/nick names
    $first_name = get_user_meta( $user_id, 'first_name', true );
    $last_name = get_user_meta( $user_id, 'last_name', true );
    $nickname = get_user_meta( $user_id, 'nickname', true );

    $contract_name = trim( $first_name . ' ' . $last_name );

    // use nickname if no first/last names
    if ( empty( $contract_name ) ) {
      $contract_name = trim( $nickname );
    }
  }

  if ( empty( $contract_name ) ) {
    // fail
    $error_message = __METHOD__ . ': Could not build a contract name for user ID ' . $user_id;
    error_log( $error_message );
    \App\bt_email_alert( get_option( 'blogname' ) . ' Seller Issue', $error_message );

    $contract_name = 'USER ID ' . $user_id;
  }

  if ( $sanitized ) {
    return sanitize_title( $contract_name );
  } else {
    return $contract_name;
  }
}


/**
 * Determine if seller has agreed to contract. Returns array of agreement data if so.
 *
 * @param null $user_id
 *
 * @return mixed False or empty value if no yet agreed. Array of agreement data if so, like this:
 *
 * [date] => 2020-05-09 14:49:54
 * [name_entered] => Test artist 0429a
 * [name_system] => Test Artist 0429a
 * [contract_text] => <p>Content goes here&#8230;</p>
 *
 */
function bt_seller_has_agreed_to_contract( $user_id = null ) {
  if ( empty( $user_id ) ) {
    $user_id = get_current_user_id();
  }

  if ( empty( $user_id ) ) {
    return false;
  }

  // get uploaded file from user meta
  $agreement = get_user_meta( $user_id, '_contract_agreement', true );

  if ( empty( $agreement ) ) {
    return false;
  }

  //error_log( __METHOD__ . ': agreement = ' . print_r( $agreement, 1 ) );

  return $agreement;
}


/**
 * Determine if seller has agreed to manual payouts.
 *
 * @param null $user_id
 *
 * @return bool
 */
function bt_seller_has_agreed_to_manual_payouts( $user_id = null ) {
  if ( empty( $user_id ) ) {
    $user_id = get_current_user_id();
  }

  if ( empty( $user_id ) ) {
    return false;
  }

  if ( empty( get_user_meta( $user_id, '_manual_payouts', true ) ) ) {
    return false;
  }

  return true;
}


/**
 * AJAX action to process agreement of manual payouts. Saves user meta upon success.
 */
add_action( 'wp_ajax_bt_agree_to_manual_payouts', 'bt_agree_to_manual_payouts' );
//add_action( 'wp_ajax_nopriv_bt_agree_to_manual_payouts', 'bt_agree_to_manual_payouts' );
function bt_agree_to_manual_payouts() {
  $data = new \stdClass();
  $data->success = false;
  $data->message = '';
  $data->message_html = '';

  //error_log( 'AJAX FOR MANUAL PAYOUT ACCEPTANCE: request = ' . print_r( $_REQUEST, 1 ) );

  if ( ! bt_is_seller() ) {
    $data->message = 'Error: user is not a seller.';

  } else {
    // entered name matches system name
    $data->success = true;
    $data->message = 'Success!';

    //error_log( 'AJAX FOR MANUAL PAYOUT ACCEPTANCE: success! user id = ' . get_current_user_id() );

    update_user_meta( get_current_user_id(), '_manual_payouts', 1 );

    // approve seller if all criteria are now met
    bt_approve_seller( get_current_user_id() );
  }

  // build html result message
  if ( $data->success ) {
    $data->message_html = '<p class="alert alert-success">' . $data->message . '</p>';
  } else {
    $data->message_html = '<p class="alert alert-danger">' . $data->message . '</p>';
  }

  // return data as JSON string
  echo json_encode( $data );
  wp_die();
}


/**
 * AJAX action to check the seller requirements and construct html to report back. This is shown with the Admin
 * "Seller is Approved" input, so that it's easy to know the status.
 */
add_action( 'wp_ajax_bt_seller_requirements', 'bt_seller_requirements' );
//add_action( 'wp_ajax_nopriv_bt_seller_requirements', 'bt_seller_requirements' );
function bt_seller_requirements() {
  $result = '';

  //error_log( __METHOD__ . ': request = ' . print_r( $_REQUEST, 1 ) );

  if ( ! isset( $_REQUEST['post_id'] ) || empty( $_REQUEST['post_id'] ) ) {
    $result = 'Error: no seller post ID provided!';
  }

  // get seller post and its author/user
  $seller_post = get_post( $_REQUEST['post_id'] );
  $user_id = $seller_post->post_author;

  // contract
  if ( empty( $agreement = bt_seller_has_agreed_to_contract( $user_id ) ) ) {
    $result .= '<p class="no">Contract: Not yet.</p>';
  } else {
    $result .= '<p class="yes">Contract: Accepted ' . $agreement['date'] . ' with name ' . $agreement['name_entered'] . '.</p>';
  }

  // sales tax registration (or alt file if set in theme-supports/config)
  $label = 'Sales Tax Registration';
  if ( $seller_sales_tax_registration_alt_file_label = bt_use_sellers( 'seller_sales_tax_registration_alt_file_label' ) ) {
    $label = $seller_sales_tax_registration_alt_file_label;
  }
  if ( empty( $file = bt_seller_has_sales_tax_registration( $user_id ) ) ) {
    $result .= '<p class="no">' . $label . ': Not yet.</p>';
  } else {
    $result .= '<p class="yes">' . $label . ': <a href="' . $file['url'] . '" target="_blank">' . $file['filename'] . '</a>.</p>';
  }

  // stripe connection
  if ( empty( $stripe_account_id = bt_get_user_stripe_account_id( $user_id ) ) ) {
    $result .= '<p class="no">Stripe Connection: Not yet.</p>';
    // manual payouts - checking this when no Stripe connection, from when we had to do manual payouts while Stripe was delayed
    if ( ! bt_seller_has_agreed_to_manual_payouts( $user_id ) ) {
      $result .= '<p class="no">Manual Payouts: No.</p>';
    } else {
      $result .= '<p class="yes">Manual Payouts: Yes.</p>';
    }

  } else {
    $result .= '<p class="yes">Stripe Connection: account ' . $stripe_account_id . '.</p>';
  }

  echo '<div class="seller-requirements">' . $result . '</div>';
  wp_die();
}


/**
 * AJAX action to process agreement of seller contract. Will make sure a date was provided (automatic) and a name was
 * entered by user, and that the name on file (in the system) matches the name entered (sanitized). Saves user meta
 * upon success.
 */
add_action( 'wp_ajax_bt_agree_to_seller_contract', 'bt_agree_to_seller_contract' );
//add_action( 'wp_ajax_nopriv_bt_agree_to_seller_contract', 'bt_agree_to_seller_contract' );
function bt_agree_to_seller_contract() {
  $data = new \stdClass();
  $data->success = false;
  $data->message = '';
  $data->message_html = '';

  //error_log( 'AJAX FOR SELLER CONTRACT: request = ' . print_r( $_REQUEST, 1 ) );

  if ( ! isset( $_REQUEST['contract_name_entered'] ) || ! isset( $_REQUEST['contract_date'] ) ) {
    $data->message = 'Error: name or date was not provided.';

  } else if ( ! bt_is_seller() ) {
    $data->message = 'Error: user is not a seller.';

  } else if ( empty( $_REQUEST['contract_name_entered'] ) ) {
    $data->message = 'Please enter your name in the input field.';

  } else if ( sanitize_title( $_REQUEST['contract_name_entered'] ) != bt_seller_contract_name( null, true ) ) {
    $data->message = 'The name you entered doesn\'t match your name in our system, which appears at the top of this contract.';

  } else {
    // entered name matches system name
    $data->success = true;
    if ( $link_page_id = bt_use_sellers( 'seller_payments_setup_page_id' ) ) {
      $link_url = get_the_permalink( $link_page_id );
      $link_title = get_the_title( $link_page_id );
    } else {
      $link_url = bt_my_account_url();
      $link_title = 'My Account';
    }
    $data->message = 'Success! Thank you for selling with Curbside Marketplace! '
      . '<a href="' . $link_url . '">Back to ' . $link_title . '</a>';

    //error_log( 'AJAX FOR SELLER CONTRACT: success! entered = ' . sanitize_title( $_REQUEST['contract_name_entered'] ) . ' / system = ' . bt_seller_contract_name( null, true ) );

    // get page for seller contract, so we can get its contents for recordkeeping
    $contract_content = '';
    if ( ! empty( $contract_page = get_post( bt_use_sellers( 'seller_contract_page_id' ) ) ) ) {
      $contract_content = apply_filters( 'the_content', $contract_page->post_content );
    }

    // build data to save in user meta
    $contract_data = [
      'date' => $_REQUEST['contract_date'],
      'name_entered' => $_REQUEST['contract_name_entered'],
      'name_system' => bt_seller_contract_name(),
      'contract_text' => bt_str_replace_contract_text( $contract_content ),
    ];
    update_user_meta( get_current_user_id(), '_contract_agreement', $contract_data );

    // approve seller if all criteria are now met
    bt_approve_seller( get_current_user_id() );
  }

  // build html result message
  if ( $data->success ) {
    $data->message_html = '<p class="alert alert-success">' . $data->message . '</p>';
  } else {
    $data->message_html = '<p class="alert alert-danger">' . $data->message . '</p>';
  }

  // return data as JSON string
  echo json_encode( $data );
  wp_die();
}


/**
 * Process contract text to replace placeholders like %%PROFILE%% with dynamic information.
 *
 * @param $contract_text
 * @param int|null $user_id
 * @param \WP_Post|null $seller_post
 *
 * @return string
 */
function bt_str_replace_contract_text( $contract_text, $user_id = null, $seller_post = null ) {
  if ( empty( $seller_post ) ) {
    $seller_post = bt_get_seller_post( $user_id );
  }

  if ( empty( $seller_post ) ) {
    \App\bt_email_alert(
      get_option( 'blogname') . ' Issue',
      'In method ' . __METHOD__ . ', no seller post exists for user id ' . $user_id . '!'
    );
  } else {
    $contract_text = str_replace( '%%PROFILE%%', $seller_post->post_title, $contract_text );
  }

  return $contract_text;
}


/**
 * Determine if a seller has uploaded their sales tax registration, or whatever alternate file might be used (e.g. liquor license).
 * Just returns true if a file exists, false otherwise.
 *
 * @param null $user_id
 *
 * @return bool|array
 */
function bt_seller_has_sales_tax_registration( $user_id = null ) {
  if ( empty( $user_id ) ) {
    $user_id = get_current_user_id();
  }

  if ( empty( $user_id ) ) {
    return false;
  }

  $seller_post = bt_get_seller_post( $user_id );

  // get uploaded file from user meta (ACF field) - sales tax registration OR alt file if set in theme-supports/config
  if ( $seller_sales_tax_registration_alt_file_field = bt_use_sellers( 'seller_sales_tax_registration_alt_file_field' ) ) {
    // custom alt file instead of sales tax registration, such as liquor license
    $file = get_field( $seller_sales_tax_registration_alt_file_field, $seller_post );

  } else {
    // default sales tax registration field/file
    $file = get_field( 'sales_tax_registration', $seller_post );
  }

  //error_log( __METHOD__ . ': user id = ' . $user_id . ' / file = ' . print_r( $file, 1 ) );

  if ( empty( $file ) ) {
    return false;
  }

  return $file;
}


/**
 * Retrieve and return the processing fee rate for a given user ID. Will look for it in meta,
 * which can be set via user profile page (using ACF). If no value, will default to a rate set within.
 *
 * @param int $user_id
 *
 * @return float The decimal value such as 0.07 when the rate is 7%
 */
function bt_seller_processing_fee_rate( $user_id ) {
  // default
  $platform_fee_rate_default = 0.07;

  // get rate from user meta
  $platform_fee_rate = get_user_meta( $user_id, 'transaction_fee', true );
  //error_log( __FUNCTION__ . ' user id = ' . $user_id . ' / data from user meta: ' . $platform_fee_rate );

  // ensure we have a valid rate, which should be within the custom field bounds
  if ( ! is_numeric( $platform_fee_rate ) || $platform_fee_rate < 3.5 || $platform_fee_rate > 50 ) {
    $platform_fee_rate = $platform_fee_rate_default;
    //error_log( __FUNCTION__ . ' using default fee: ' . $platform_fee_rate );

  } else {
    $platform_fee_rate = $platform_fee_rate / 100;
    //error_log( __FUNCTION__ . ' using user-based fee: ' . $platform_fee_rate );
  }

  return $platform_fee_rate;
}


/**
 * Helper to see if a seller (post, not user) has been approved.
 *
 * @param $seller_post
 *
 * @return bool
 */
function bt_seller_is_approved( $seller_post ) {
  // check approved field on the post
  return get_field( 'seller_is_approved', $seller_post ) ? true : false;

}


/**
 * Check if seller should be approved based on criteria:
 * - contract signed
 * - sales tax registration uploaded
 * - stripe connected
 *
 * If so, set meta for approval and alert admin via email.
 *
 * @param $user_id
 */
function bt_approve_seller( $user_id ) {
  $contract = false;
  $sales_tax_registration = false;
  $stripe_connection = false;
  $manual_payouts = false;
  // $custom_requirement = false; /* could check via custom bt_use_sellers() config */

  // get seller post
  $seller_post = bt_get_seller_post( $user_id );

  // check if already approved
  $already_approved = get_field( 'seller_is_approved', $seller_post ) ? true : false;
  if ( $already_approved ) {
    return;
  }

  // contract
  if ( ! empty( $agreement = bt_seller_has_agreed_to_contract( $user_id ) ) ) {
    $contract = true;
  }

  // sales tax registration
  if ( ! empty( $file = bt_seller_has_sales_tax_registration( $user_id ) ) ) {
    $sales_tax_registration = true;
  } /* else if ( not required via bt_use_sellers() ) { true } */

  // stripe connection
  if ( ! empty( $stripe_account_id = bt_get_user_stripe_account_id( $user_id ) ) ) {
    $stripe_connection = true;
  }

  // manual payouts
  if ( bt_seller_has_agreed_to_manual_payouts() ) {
    $manual_payouts = true;
  }

  // approve seller if all requirements are met
  if ( $contract
    && $sales_tax_registration
    && ( $stripe_connection || $manual_payouts )
  ) {

    // get user data
    $user_name = bt_seller_contract_name( $user_id );
    $user_obj = get_userdata( $user_id );
    $user_email = $user_obj->user_email;

    // set post meta for seller being approved
    update_field( 'seller_is_approved', '1', $seller_post );

    // determine edit post and user URLs for admins to reference
    $edit_post_url = get_admin_url() . 'post.php?post=' . $seller_post->ID . '&action=edit';
    $edit_user_url = get_admin_url() . 'user-edit.php?user_id=' . $user_id;

    // build email to alert about approval
    $subject = 'Seller Approved on ' . get_bloginfo( 'name' );
    $message = <<<MSG
The following {$seller_post->post_type} has completed all steps to be approved for online sales, and has been approved.

Please review and make sure all is well.

Post: {$seller_post->post_title}
$edit_post_url

User: $user_name
$edit_user_url

Email: $user_email
MSG;

    $to = [
      'Curbside Marketplace <hello@curbsidemarketplace.com>',
    ];
    $headers = [
      'From: ' . get_bloginfo( 'name' ) . ' Website <no-reply@curbsidemarketplace.com>',
    ];

    // send the email
    if ( defined( 'BT_ENV') && BT_ENV == 'dev' ) {
      // just log if dev env
      error_log( 'SEND APPROVAL EMAIL: ' . $subject . ' / ' . $message );
    } else {
      wp_mail( $to, $subject, $message, $headers );
      error_log( 'SENT APPROVAL EMAIL: ' . $subject . ' / ' . $message );
    }
  }
}


/**
 * AJAX responder to build and output a weekly payouts report.
 *
 * This is shown in wp-admin (on seller post) for administrators and marketplace admins,
 * and on my-account page for sellers. For admins (but not marketplace admins) in
 * wp-admin, upcoming payouts include a trigger to do/record the payout (manual or stripe based on user data).
 *
 * Each week (starts Monday, thru Sunday) is a row in the report, and can be expanded to show order detail.
 *
 * Totals, fees, tips, etc. are calculated on the fly, but if a payout has already been made for a week, the fee,
 * fee rate, paid amount, and any adjustment comes from the stored payout data.
 *
 * For weeks that haven't been paid, if there is any carryover difference from prior/paid weeks (e.g. from orders that
 * are refunded after they've been paid out) then that will be accounted for as an adjustment in upcoming payout(s).
 */
add_action( 'wp_ajax_bt_payouts_report', 'bt_payouts_report' );
//add_action( 'wp_ajax_nopriv_bt_payouts_report', 'bt_payouts_report' );
function bt_payouts_report() {

  // verify request
  check_ajax_referer( 'bt_ecomm_nonce', 'nonce' );

  // determine which user we are getting data for
  $user_id = 0;
  if ( wc_current_user_has_role( 'shop_manager' )
    && ! wc_current_user_has_role( 'administrator' )
    && ! wc_current_user_has_role( 'marketplace_administrator' )
  ) {
    // seller - current user
    $user_id = get_current_user_id();

  } else if ( bt_can_admin_marketplace() && isset( $_REQUEST['post_id'] ) && $_REQUEST['post_id'] > 0 ) {
    // admin - use the current post being edited to determine its seller
    $user_id = bt_get_author_id_for_post( $_REQUEST['post_id'] );

  } else {
    wp_die( 'Cannot determine which user to get reports for. If this is for a not-yet-created new user, that is why and it\'s ok.' );
  }

  // init vars for weekly data, grand totals, and seller fee rate
  $weeks = [];
  $orders_grand_total = 0;
  $orders_net_grand_total = 0;
  $platform_fee_grand_total = 0;
  $orders_count_total = 0;
  $tips_grand_total = 0;
  $platform_fee_rate_current = bt_seller_processing_fee_rate( $user_id );
  $stripe_account_id = bt_get_user_stripe_account_id( $user_id );

  // init current date vars
  $current_week = (int) wp_date( 'W' );
  $current_year = (int) wp_date( 'o' );

  // calculate start of week
  // NOTE: with cotc, there were issues because week started on Sunday instead of Monday (as aaw does). USE MONDAY.
  $start_of_week = get_option( 'start_of_week', 1 );
  //error_log( __METHOD__ . ': start of week = ' . $start_of_week );

  // determine the date vars for when orders start (the oldest orders)
  // -- can't use wc_get_orders because even with post_author it will return oldest order of all when no orders for author/user
  $oldest_order = new \WP_Query( [
    'post_type' => 'shop_order',
    'author' => $user_id,
    'posts_per_page' => 1,
    'post_status' => [
      'wc-completed',
      'wc-completed-picked',
      'wc-completed-shipped',
      // 'wc-refunded', ==> we'll only get completed, so that needed refunds bubble up to next payout
    ],
    'orderby' => 'date',
    'order' => 'ASC',
  ] );
  $oldest_order = $oldest_order->posts;

  //echo( '<pre>' . print_r( $_REQUEST, 1 ) . '</pre>' );
  //wp_die( '<pre>' . print_r( $oldest_order, 1 ) . '</pre>' );

  // bail if there aren't any orders yet
  if ( empty( $oldest_order ) ) {
    wp_die( '<p class="mb-0">Once you have completed an order, data will appear here.</p>' );
  }

  // grab the order out of the array returned
  $oldest_order = new \WC_Order( current( $oldest_order )->ID );

  // calculate oldest order time
  $oldest_order_time = strtotime( $oldest_order->get_date_created() );
  //error_log( __METHOD__ . ': oldest order created date = ' . $oldest_order_time );
  $orders_start_week = (int) wp_date( 'W', $oldest_order_time );
  $orders_start_year = (int) wp_date( 'o', $oldest_order_time );
  //error_log( __METHOD__ . ': START WEEK = ' . $orders_start_week );
  //error_log( __METHOD__ . ': START YEAR = ' . $orders_start_year );

  // emergency fallback
  if ( ! is_numeric( $orders_start_week ) || ! is_numeric( $orders_start_year ) ) {
    $orders_start_week = 14;
    $orders_start_year = 2020;
  }

  /* get current seller payout data - is array of objects like:
    [
      '2020W18' => stdClass Object (
        'date_paid' => '2020-04-30 08:08:08',
        'amount' => '123.45', // payout amount net (adjustment is already accounted for)
        'adjustment' => '-12.34', // to account for refunds that happen after the order has been paid out
        'fee_rate' => 0.07,
        'admin_id' => 1,
        'stripe_transfer_id' => 'tr_1GmRWCJLC25qo6OyyfQCQiDB', // if stripe was used for a transfer payout, this will exist
      ),
    ]
   */
  $seller_payouts_completed = bt_seller_payouts( $user_id );
  //error_log( __METHOD__ . ': completed payouts = ' . print_r( $seller_payouts_completed, 1 ) );

  /**
   * loop weeks from current date back to when orders started, to get data in each week
   */
  //error_log( __METHOD__ . ': START WEEK / YEAR = ' . $orders_start_week . ' / ' . $orders_start_year );
  //error_log( __METHOD__ . ': CURRENT WEEK / YEAR = ' . $current_week . ' / ' . $current_year );
  while (
    // in a year after the start year, so keep going
    $current_year > $orders_start_year
    || (
      // in the start year, and a week on or after the start year, so keep going
      $current_year == $orders_start_year
      && $current_week >= $orders_start_week
    )
  ) {
    //error_log( __METHOD__ . ': PROCESSING WEEK / YEAR = ' . $current_week . ' / ' . $current_year );

    // init vars for weekly totals
    $seller_orders_total = 0;
    $platform_fee_total = 0;
    $tips_total = 0;

    // query for completed & refunded orders during this week
    /**
     * @todo Resolve issues with orders placed on the first days of a new year that technically fall into week 52 of the prior year!
     * In 2023, there was an order placed for a seller on 1/1, which was caught in the reports for payouts to do, but
     * wasn't in the set of orders to pay out on seller edit screen in wp-admin. The query below doesn't return an order
     * like that with W52 or W53. Maybe looking into the payouts-to-do report would reveal how we can better query.
     *
     * For now, the resolution was to change the order date to the first day of the first full week in the current year.
     *
     * Testing with modified sql queries generated from below, using e.g. 2023 W0 would return the order in question!
     *
     * We tried changing
     *   if ( $current_week <= 0 )
     * below to
     *   if ( $current_week < 0 )
     * but that didn't work and there didn't seem to be an easy adjustment to the current logic to make it work.
     */
    $wp_query_orders = new \WP_Query( [
      'post_type' => 'shop_order',
      'author' => $user_id,
      'nopaging' => true,
      'post_status' => [
        'wc-completed',
        'wc-completed-picked',
        'wc-completed-shipped',
        // 'wc-refunded', ==> we'll only get completed, so that needed refunds bubble up to next payout
      ],
      'orderby' => 'date',
      'order' => 'DESC',
      'date_query' => [
        'relation' => 'AND',
        'week' => $current_week,
        'year' => $current_year,
      ],
    ] );

    //echo '<pre>' . print_r( $wp_query_orders, 1 ) . '</pre>';

    // get padded current_week, since a leading zero is expected for single digit weeks
    $current_week_padded = str_pad( (string)$current_week, 2, '0', STR_PAD_LEFT );

    // slug for current year+week, used as key in week and payout data arrays
    $year_week = $current_year . 'W' . $current_week_padded;
    // determine when a week starts and ends based on the site's option for the day a week starts on
    $year_week_start = $current_year . 'W' . $current_week_padded . ( $start_of_week ? ' +' . $start_of_week . ' days' : '' );
    $year_week_end = $current_year . 'W' . $current_week_padded . ' +' . (6 + $start_of_week) . ' days';
    //error_log( __METHOD__ . ' -- year week start & end = ' . $year_week_start . ' - ' . $year_week_end );
    // init this week's data
    $week_data = [
      'week' => $current_week,
      'year' => $current_year,
      'first_day' => wp_date( 'Y-m-d', strtotime( $year_week_start ) ),
      'last_day' => wp_date( 'Y-m-d', strtotime( $year_week_end ) ),
      'order_count' => $wp_query_orders->post_count,
      'orders' => [],
      'payout_adjustment' => 0, // this may get changed later when checking for diffs e.g. refund after order has been paid out to seller
    ];
    //echo '<pre>$week_data for ' . $year_week . ' = ' . print_r( $week_data, 1 ) . '</pre>';
    $orders_count_total += $wp_query_orders->post_count;

    // determine what fee rate is used for the current week - uses payout data if paid, else the current rate
    if ( array_key_exists( $year_week, $seller_payouts_completed ) ) {
      // use fee rate from the existing payout
      $fee_rate_to_use = $seller_payouts_completed[$year_week]->fee_rate;
    } else {
      // use current user fee rate
      $fee_rate_to_use = $platform_fee_rate_current;
    }

    // loop orders to calculate their fees, net amounts, weekly totals, etc.
    foreach ( $wp_query_orders->posts as $order ) {
      // make an order object and get the total
      $order_post = new \WC_Order( $order->ID );
      $order_total = (float) $order_post->get_total();

      // check for refunds
      $refunded_amount = (float) $order_post->get_total_refunded();
      if ( $refunded_amount ) {
        /* * /
        error_log( __METHOD__ . ': refunded amount = ' . $refunded_amount
                   . ' / order = ' . $order->ID
                   . ' / original total = ' . $order_total
                   . ' / new total = ' . ( $order_total - $refunded_amount )
        );
        / * */
        $order_total -= $refunded_amount;
      }

      // calculate platform fee for this order
      $platform_fee = round( $fee_rate_to_use * $order_total, 2 );

      // add this order's total and platform fee to the seller's grand total
      $seller_orders_total += $order_total;
      $platform_fee_total += $platform_fee;

      $tip_amount = 0;

      // find our tip fee
      foreach ( $order_post->get_fees() as $fee ) {
        if ( strpos( $fee->get_name(), 'Tip' ) !== false ) {
          // get amount
          $tip_amount = $fee->get_total();
          // add total tips
          $tips_total += floatval( $tip_amount );
          break;
        }
      }

      // set dates
      $date_placed = $order_post->get_date_created();
      if ( is_object( $date_placed ) ) {
        $date_placed = $date_placed->date( 'M j g:ia' );
      }
      $date_completed = $order_post->get_date_completed();
      if ( is_object( $date_completed ) ) {
        $date_completed = $date_completed->date( 'M j g:ia' );
      }

      // add this order to the week's data
      $week_data['orders'][$order->ID] = [
        'ID' => $order->ID,
        'date_placed' => $date_placed,
        'date_completed' => $date_completed,
        'total' => $order_total,
        'platform_fee' => $platform_fee,
        'net' => $order_total - $platform_fee,
        'net_less_tip' => $order_total - $platform_fee - $tip_amount,
        'tip_amount' => $tip_amount,
        //'tip_amount_html' => '$' . number_format( $tip_amount, 2 ),
        //'tip_percent_html' => trim( str_replace( 'Tip', '', $fee->get_name() ) ),
        //'WP_Post' => $order,
      ];
    }

    // calculate seller net total to be transferred
    $seller_net = $seller_orders_total - $platform_fee_total;

    // add totals for this week to seller data and grand totals
    // - total
    $week_data['orders_total'] = $seller_orders_total;
    $orders_grand_total += $seller_orders_total;
    // - platform fee
    $week_data['platform_fee_total'] = $platform_fee_total;
    $platform_fee_grand_total += $platform_fee_total;
    // - net
    $week_data['orders_net_total'] = $seller_net;
    $orders_net_grand_total += $seller_net;
    // - tips
    $week_data['tips_total'] = $tips_total;
    $tips_grand_total += $tips_total;

    // add this week's data to the set of all weeks
    $weeks[$year_week] = $week_data;

    // decrement our time periods
    $current_week--;
    if ( $current_week <= 0 ) {
      // into the prior year
      $current_year--;
      $current_week = (int) wp_date( 'W', strtotime( 'December 31, ' . $current_year ) );
      //error_log( __METHOD__ . ' -- JUST LOOPED CURRENT YEAR / WEEK TO = ' . $current_year . ' / ' . $current_week );
    }
  }

  /**
   * loop the weeks in reverse to calculate adjustments (e.g. if refund after payout) and carryover balance.
   *
   * init var for difference carryover, and for whether we've accounted for a positive adjustment (e.g. orders
   * completed after their week was paid out), so that we only do this once (the first/current week to be paid out)
   *
   * A week looks like: Array
  (
  [week] => 37
  [year] => 2020
  [first_day] => 2020-09-07
  [last_day] => 2020-09-13
  [order_count] => 0
  [orders] => Array(...)
  [payout_adjustment] => 0
  [orders_total] => 0
  [platform_fee_total] => 0
  [orders_net_total] => 0
  [tips_total] => 0
  )
   */
  $difference_carryover = 0;
  $accounted_for_positive_adjustment = false;

  $counter = 0;
  $num_weeks = count( $weeks );
  $current_payout_week = '';
  foreach ( array_reverse( $weeks, true ) as $year_week => $week ) {
    $counter++;

    if ( $counter == $num_weeks - 1 ) {
      // the second-to-last week in the array is the current week to pay out
      $current_payout_week = $year_week;
      //error_log( __METHOD__ . ' -- current payout week = ' . print_r( $current_payout_week, true ) );
    }

    // ensure difference carryover isn't a number like -5.3290705182008E-15
    $difference_carryover = round( $difference_carryover, 2 );

    // bool for whether this week's payout has been made already
    $week_payout_made = array_key_exists( $year_week, $seller_payouts_completed );

    // determine how much has been paid out
    if ( $week_payout_made ) {
      $paid = $seller_payouts_completed[$year_week]->amount;
    } else {
      $paid = 0;
    }

    /* * /
      error_log(
          __METHOD__ . ': year_week = ' . $year_week
             . ' / week payout made = ' . ( $week_payout_made ? 'TRUE' : 'FALSE' )
             . ' / paid = ' . $paid
             . ' / difference carryover = ' . $difference_carryover
             . ' / week orders net total = ' . $week['orders_net_total']
      );
    / * */

    // process carryover from a prior week, so long as a payout hasn't already been made this week, AND the week
    // has something to pay out (otherwise we flip the $accounted_for_positive_adjustment bool if there's a week
    // without sales)
    if ( ! $week_payout_made && $week['orders_net_total'] > 0 ) {

      if ( $difference_carryover < 0 ) {
        // negative carryover - e.g. a refund happened for an order that was already paid out
        if ( $difference_carryover <= $week['orders_net_total'] ) {
          // there's enough in this week's net/payout to deduct the full carryout from it,
          // so adjust payout by full carryover amount and reset carryover
          $weeks[$year_week]['payout_adjustment'] += $difference_carryover;
          $difference_carryover = 0;
        } else {
          // the carryover is more than this week's net/payout,
          // so adjust payout by by this week's net and subtract that much from the carryover
          // to leave remainder
          $weeks[$year_week]['payout_adjustment'] += $week['orders_net_total'];
          $difference_carryover -= $week['orders_net_total'];
        }

      } else if ( $difference_carryover > 0 && ! $accounted_for_positive_adjustment ) {
        // positive carryover - e.g. a past order was completed after that week was paid out,
        // so increase this week's net/payout by the difference
        $weeks[ $year_week ]['payout_adjustment'] += $difference_carryover;
        $difference_carryover = 0;
        // set the bool so we only do this once
        $accounted_for_positive_adjustment = true;

      } else {
        // zero carryover - set the bool for positive adjustments being accounted for, since we only
        // want that to come into play when there's a positive carryover the first week without a
        // payout (otherwise they stack up into the following weeks if there are more)
        $accounted_for_positive_adjustment = true;
      }

    } else if (
      $year_week == $current_payout_week
      && empty( $week['orders_net_total'] )
      && $difference_carryover > 0
      && ! $accounted_for_positive_adjustment
    ) {
      // it's the payout week, with zero orders, but there's a positive adjustment that hasn't been accounted for.
      // e.g. an order was completed in an week already paid out, but no other sales to piggyback adjustment onto.
      $weeks[ $year_week ]['payout_adjustment'] += $difference_carryover;
      $difference_carryover = 0;
      // set the bool so we only do this once
      $accounted_for_positive_adjustment = true;

    } else if ( $week_payout_made ) {
      // payout has been made, include any adjustment data from it
      // and account for the adjustment in carryover and paid amounts
      $weeks[$year_week]['payout_adjustment'] = $seller_payouts_completed[$year_week]->adjustment;
      $difference_carryover -= $seller_payouts_completed[$year_week]->adjustment;
      // and we account for the adjustment in the paid amount so that the difference is correct
      $paid -= $seller_payouts_completed[$year_week]->adjustment;
    }

    // make sure payout adjustment isn't a number like -0.00000000000000532907
    $weeks[$year_week]['payout_adjustment'] = round( $weeks[$year_week]['payout_adjustment'], 2 );

    // calculate difference and what's owed
    $week_difference = round( $week['orders_net_total'] - $paid, 2 );
    $weeks[$year_week]['difference'] = $week_difference;

    // accumulated difference
    $difference_carryover += $week_difference;

    //error_log(__METHOD__ . ': user ' . $user_id . ' / week ' . $year_week . ' / net ' . $week['orders_net_total'] . ' / paid ' . $paid . ' / diff ' . $week_difference . ' / carryover ' . $difference_carryover );
  }

  //echo 'Hello. Current date W o: ' . wp_date( 'W o' );
  //echo '<pre>' . print_r( $weeks, 1 ) . '</pre>';
  ?>
  <table>
    <thead>
    <tr>
      <th class="text">Week</th>
      <th>Total</th>
      <th>Platform Fee</th>
      <th>Net</th>
      <th class="ok">Payout</th>
      <th class="ok">Owed</th>
      <?php if ( bt_use_tips() ) : ?>
        <th class="minor">Tips</th>
        <th class="minor">Net Less Tips</th>
      <?php endif; ?>
    </tr>
    </thead>
    <tbody>
    <?php foreach ( $weeks as $year_week => $week ) :
      /* * /
    $week_to_log = $week;
    $week_to_log['orders'] = '...count = ' . count( $week['orders'] );
    error_log( __METHOD__ . ' -- week of ' . $year_week . ' = ' . print_r( $week_to_log, true ) );
    / * */

      // bool for whether this week's payout has been made already
      $week_payout_made = array_key_exists( $year_week, $seller_payouts_completed );
      ?>
      <tr class="week">
        <td class="text">
          <?= $week['order_count'] ? '<a class="toggle" href="' . $year_week . '" data-year_week="' . $year_week . '" data-open="0">' : '' ?>
          <?= $week['first_day'] ?>
          <?= $week['order_count'] ? '</a>' : '' ?>
        </td>
        <td>
          $<?= number_format( $week['orders_total'], 2 ) ?>
          <div class="minor small nowrap">
            <?= number_format( $week['order_count'] ) . ' order' . ( $week['order_count'] == 1 ? '' : 's' ) ?>
          </div>
        </td>
        <td>
          -$<?= number_format( $week['platform_fee_total'], 2 ) ?>
          <div class="minor small">
            <?= $week_payout_made
              ? $seller_payouts_completed[$year_week]->fee_rate * 100
              : $platform_fee_rate_current * 100
            ?>%
          </div>
        </td>
        <td>
          $<?= number_format( $week['orders_net_total'], 2 ) ?>
          <?= $week['payout_adjustment'] != 0
            ? '<div class="' . ( $week['payout_adjustment'] > 0 ? 'ok' : 'warn' ) . ' small">$' . number_format( $week['payout_adjustment'], 2 ) . ' adj</div>'
            : '' ?>
        </td>

        <?php if ( $week_payout_made ) : ?>
          <td class="ok">
            $<?= number_format( $seller_payouts_completed[$year_week]->amount, 2 ) ?>
            <div class="minor small nowrap">
              sent <?= wp_date( 'n/j', strtotime( $seller_payouts_completed[$year_week]->date_paid ) ) ?>
              <?= $seller_payouts_completed[$year_week]->stripe_transfer_id
                ? ' via <span title="' . $seller_payouts_completed[$year_week]->stripe_transfer_id . '">Stripe</span>'
                : ' manually' ?>
            </div>
          </td>
          <td class="<?= $week['difference'] == 0 ? 'minor' : 'warn' ?>">
            $<?= number_format( $week['difference'], 2 ) ?>
            <?= $week['difference'] < 0
              ? '<div class="minor small">refund adj</div>'
              : ''
            ?>
          </td>
        <?php else : // payout NOT made yet
          $payout_amount = $week['orders_net_total'] + $week['payout_adjustment'];
          ?>
          <?php if ( wc_current_user_has_role( 'administrator' ) ) : ?>
          <td>
            <?php if ( $payout_amount > 0 && bt_week_can_be_paid( $year_week ) ) : ?>
              <a class="do-payout"
                 data-user_id="<?= $user_id ?>"
                 data-year_week="<?= $year_week ?>"
                 data-amount="<?= $payout_amount ?>"
                 data-adjustment="<?= $week['payout_adjustment'] ?>"
                 data-fee_rate="<?= $platform_fee_rate_current ?>"
              >
                <?= empty( $stripe_account_id ) ? 'Record Manual Payout' : 'Do Stripe Payout' ?>
              </a>
            <?php elseif ( $payout_amount > 0 ) : ?>
              Can't Pay Out Yet
            <?php else : ?>
              None this week
            <?php endif; ?>
            <br>
            $<?= number_format( $payout_amount, 2 ) ?>
          </td>
          <td class="<?= $payout_amount > 0 ? 'warn' : 'minor' ?>">$<?= number_format( $payout_amount, 2 ) ?></td>
        <?php else : ?>
          <td class="nowrap">
            <em><?= $payout_amount > 0 ? 'Upcoming' : 'None this week' ?></em>
          </td>
          <td class="<?= $payout_amount > 0 ? 'ok' : 'minor' ?>">$<?= number_format( $payout_amount, 2 ) ?></td>
        <?php endif; ?>
        <?php endif; ?>

        <?php if ( bt_use_tips() ) : ?>
          <td class="minor">
            -$<?= number_format( $week['tips_total'], 2 ) ?>
            <?php if ( $week['tips_total'] ) : ?>
              <div class="minor small">
                <?= number_format( 100 * $week['tips_total'] / ( $week['orders_total'] - $week['tips_total'] ), 2 ) ?>%
              </div>
            <?php endif; ?>
          </td>
          <td class="minor">$<?= number_format( $week['orders_net_total'] - $week['tips_total'], 2 ) ?></td>
        <?php endif; ?>
      </tr>
      <?php if ( $week['order_count'] > 0 ) : ?>
      <tr>
        <td colspan="8" class="orders-<?= $year_week ?>">
          <div class="closed">
            <table>
              <thead>
              <tr>
                <th class="text">Order #</th>
                <th class="text">Placed</th>
                <th class="text">Completed</th>
                <th>Total</th>
                <th>Platform Fee</th>
                <th>Net</th>
                <?php if ( bt_use_tips() ) : ?>
                  <th class="minor">Tip</th>
                  <th class="minor">Net Less Tip</th>
                <?php endif; ?>
              </tr>
              </thead>
              <tbody>
              <?php foreach ( $week['orders'] as $order ) : ?>
                <tr class="order">
                  <td class="text">
                    <a href="<?= get_admin_url() . 'post.php?post=' . $order['ID'] . '&action=edit' ?>" target="_blank">
                      <?= $order['ID'] ?>
                    </a>
                  </td>
                  <td class="text"><?= $order['date_placed'] ?></td>
                  <td class="text"><?= $order['date_completed'] ?></td>
                  <td>$<?= number_format( $order['total'], 2 ) ?></td>
                  <td>-$<?= number_format( $order['platform_fee'], 2 ) ?></td>
                  <td>$<?= number_format( $order['net'], 2 ) ?></td>
                  <?php if ( bt_use_tips() ) : ?>
                    <td class="minor">-$<?= number_format( $order['tip_amount'], 2 ) ?></td>
                    <td class="minor">$<?= number_format( $order['net_less_tip'], 2 ) ?></td>
                  <?php endif; ?>
                </tr>
              <?php endforeach; ?>
              </tbody>
              <tfoot>
              <tr>
                <td class="text" colspan="3">
                  <?= count( $week['orders'] ) ?> orders &middot;
                  <?= $week['first_day'] ?> &rarr; <?= $week['last_day'] ?>
                </td>
                <td>$<?= number_format( $week['orders_total'], 2 ) ?></td>
                <td>-$<?= number_format( $week['platform_fee_total'], 2 ) ?></td>
                <td>$<?= number_format( $week['orders_net_total'], 2 ) ?></td>
                <?php if ( bt_use_tips() ) : ?>
                  <td class="minor">-$<?= number_format( $week['tips_total'], 2 ) ?></td>
                  <td class="minor">$<?= number_format( $week['orders_net_total'] - $week['tips_total'], 2 ) ?></td>
                <?php endif; ?>
              </tr>
              </tfoot>
            </table>
          </div>
        </td>
      </tr>
    <?php endif; ?>
    <?php endforeach; ?>
    </tbody>
    <tfoot>
    <tr>
      <td class="text"><?= $week_count = count( $weeks ) ?> week<?= $week_count == 1 ? '' : 's' ?></td>
      <td>
        $<?= number_format( $orders_grand_total, 2 ) ?>
        <div class="small">
          <?= number_format( $orders_count_total ) . ' order' . ( $orders_count_total == 1 ? '' : 's' ) ?>
        </div>
      </td>
      <td>-$<?= number_format( $platform_fee_grand_total, 2 ) ?></td>
      <td>$<?= number_format( $orders_net_grand_total, 2 ) ?></td>
      <td></td>
      <td></td>
      <?php if ( bt_use_tips() ) : ?>
        <td class="minor">
          -$<?= number_format( $tips_grand_total, 2 ) ?>
          <?php if ( $tips_grand_total ) : ?>
            <div class="minor small">
              <?= number_format( 100 * $tips_grand_total / ( $orders_grand_total - $tips_grand_total ), 2 ) ?>%
            </div>
          <?php endif; ?>
        </td>
        <td class="minor">$<?= number_format( $orders_net_grand_total - $tips_grand_total, 2 ) ?></td>
      <?php endif; ?>
    </tr>
    </tfoot>
  </table>
  <?php

  wp_die();
}


/**
 * AJAX responder to "do" a payout. Records data for now, but doesn't make any transactions yet.
 */
add_action( 'wp_ajax_bt_do_payout', 'bt_do_payout' );
//add_action( 'wp_ajax_nopriv_bt_do_payout', 'bt_do_payout' );
function bt_do_payout() {
  //error_log( __METHOD__ . ': request = ' . print_r( $_REQUEST, 1 ) );

  // verify request
  check_ajax_referer( 'bt_ecomm_nonce', 'nonce' );

  // sanity checks

  if ( ! function_exists( 'wc_current_user_has_role' ) || ! wc_current_user_has_role( 'administrator' ) ) {
    echo 'Error: you are not an administrator.';
    wp_die();
  }

  if ( ! isset( $_REQUEST['user_id'] ) || ! is_numeric( $_REQUEST['user_id'] ) ) {
    echo 'Error: no User ID provided.';
    wp_die();
  }

  if ( ! isset( $_REQUEST['year_week'] ) ) {
    echo 'Error: no year-week provided.';
    wp_die();
  }

  if ( ! isset( $_REQUEST['amount'] ) || ! is_numeric( $_REQUEST['amount'] ) ) {
    echo 'Error: no amount provided.';
    wp_die();
  }

  if ( ! isset( $_REQUEST['adjustment'] ) || ! is_numeric( $_REQUEST['adjustment'] ) ) {
    echo 'Error: no adjustment provided.';
    wp_die();
  }

  if ( ! isset( $_REQUEST['fee_rate'] ) || ! is_numeric( $_REQUEST['fee_rate'] ) ) {
    echo 'Error: no fee rate provided.';
    wp_die();
  }

  // get stripe account ID so we know whether we're doing stripe transfers or manual payouts
  $stripe_account_id = bt_get_user_stripe_account_id( $_REQUEST['user_id'] );

  // get existing payout data
  $seller_payouts_completed = bt_seller_payouts( $_REQUEST['user_id'] );

  // can't do a payout for a week we've already done
  if ( array_key_exists( $_REQUEST['year_week'], $seller_payouts_completed ) ) {
    echo 'Error: year-week already has a payout.';
    wp_die();
  }

  // if stripe is connected (we have account ID) do the transfer
  if ( ! empty( $stripe_account_id ) ) {
    $seller_net_no_decimal = round( $_REQUEST['amount'] * 100 );

    // perform a transfer
    if ( $seller_net_no_decimal > 0 ) {

      // set our secret key for working with stripe
      \Stripe\Stripe::setApiKey( STRIPE_CONNECT_KEY_SECRET );

      // check for available balance
      try {
        $balance = \Stripe\Balance::retrieve();
        $balance_amount = current( $balance->available )->amount;
        //error_log( __FUNCTION__ . ' balance: ' . print_r( $balance, 1 ) );
        //error_log( __FUNCTION__ . ' balance amount: ' . print_r( $balance_amount, 1 ) );

        if ( $seller_net_no_decimal > $balance_amount ) {
          $message = 'Amount to pay out (' . $seller_net_no_decimal . ') is greater than available balance (' . $balance_amount . ')!';
          error_log( __FUNCTION__ . ': ' . $message );
          \App\bt_email_alert( get_option( 'blogname' ) . ': Error processing seller payout', $message );
          wp_die( $message );
        }

      } catch ( \Exception $e ) {
        error_log( __FUNCTION__ . ' exception: ' . print_r( $e->getError(), 1 ) );
        $message = 'An unknown error occurred while checking available balance. Check error log for details.';
        \App\bt_email_alert( get_option( 'blogname' ) . ': Error checking Stripe balance', $message );
        wp_die( $message );
      }

      // make the transfer
      try {
        $transfer = \Stripe\Transfer::create([
          'amount' => $seller_net_no_decimal,
          'currency' => 'usd',
          'destination' => $stripe_account_id,
          // ONLY USE WHEN CHARGE TRANSACTIONS MADE WITH THIS, RIGHT? 'transfer_group' => $transfer_group,
        ]);
      } catch ( \Exception $e ) {
        error_log( __FUNCTION__ . ' exception: ' . print_r( $e->getError(), 1 ) );
        $message = 'An unknown error occurred while making a transfer for seller payout. Check error log for details.';
        \App\bt_email_alert( get_option( 'blogname' ) . ': Error processing seller payout', $message );
        wp_die( $message );
      }
    }

  }

  // record payout data in our custom table
  global $wpdb;
  $wpdb->insert(
    $wpdb->prefix . 'bt_payouts',
    [
      'user_id' => $_REQUEST['user_id'],
      'admin_id' => get_current_user_id(),
      'year_week' => $_REQUEST['year_week'],
      'date_paid' => wp_date( 'Y-m-d H:i:s' ),
      'amount' => $_REQUEST['amount'],
      'adjustment' => $_REQUEST['adjustment'],
      'fee_rate' => $_REQUEST['fee_rate'],
      'stripe_transfer_id' => ( isset( $transfer ) ? $transfer->id : '' ),
    ]
  );

  // email alert to seller
  $payout_amount_formatted = '$' . number_format( $_REQUEST['amount'], 2 );
  $payout_week_datetime = new \DateTime( $_REQUEST['year_week'], new \DateTimeZone( get_option( 'timezone_string' ) ) );
  $payout_week_formatted = $payout_week_datetime->format( 'n/j' )
    . ' - ' . $payout_week_datetime->add( new \DateInterval( 'P6D' ) )->format( 'n/j' );
  $payout_method_message = isset( $transfer )
    ? 'This amount is being transferred to your bank account via Stripe (transfer ID ' . $transfer->id . ').'
    : 'A check for this amount is on its way to you.';
  $my_account_url = bt_my_account_url();

  // get user data
  $user_name = bt_seller_contract_name( $_REQUEST['user_id'] );
  $user_obj = get_userdata( $_REQUEST['user_id'] );
  $user_email = $user_obj->user_email;

  // build email to alert about approval
  $subject = get_bloginfo( 'name' ) . ' Payout for ' . $user_name;
  $message = <<<MSG
Hello {$user_name} -

A payout of {$payout_amount_formatted} has been initiated for completed orders in the week of {$payout_week_formatted}.

{$payout_method_message}

If any orders in this period change after this payout (e.g. a refund is issued, or an order that was still "processing" is "completed"), the difference will flow into your next payout.

See reports for your weekly sales and payouts within My Account:
{$my_account_url}

If you have questions, simply reply to this email.

Thank you!
- Curbside Marketplace
MSG;

  $to = [
    $user_name . ' <' . $user_email . '>',
  ];
  $headers = [
    'From: Curbside Marketplace <hello@curbsidemarketplace.com>',
    'Bcc: Curbside Marketplace <' . get_bloginfo( 'admin_email' ) . '>',
  ];

  // send the email
  if ( defined( 'BT_ENV') && BT_ENV == 'dev' ) {
    // just log if dev env
    error_log( 'SEND PAYOUT EMAIL: to ' . $user_email . ' / ' . $subject . ' / ' . $message . ' / headers = ' . print_r( $headers, 1 ) );
  } else {
    wp_mail( $to, $subject, $message, $headers );
    error_log( 'SENT PAYOUT EMAIL: ' . $subject . ' / ' . $message );
  }

  // success message
  echo 'Recorded payment of $' . number_format( $_REQUEST['amount'], 2 ) . ' successfully.';
  if ( isset( $transfer ) ) {
    echo ' Stripe transfer ID ' . $transfer->id;
  }
  wp_die();
}


/**
 * Retrieve records of payouts for a seller user.
 *
 * @param int|null $user_id
 *
 * @return array
 */
function bt_seller_payouts( $user_id = null ) {
  if ( empty( $user_id ) ) {
    $user_id = get_current_user_id();
  }

  global $wpdb;

  // build sql to get payout records for this seller
  $sql = 'SELECT `year_week`, `date_paid`, `amount`, `adjustment`, `fee_rate`, `stripe_transfer_id`
			FROM `' . $wpdb->prefix . 'bt_payouts`
			WHERE `user_id` = ' . $user_id . '
			ORDER BY `date_paid` DESC';

  //error_log( __METHOD__ . ': sql = ' . $sql );

  return $wpdb->get_results( $sql, OBJECT_K );
}


/**
 * @param $year_week string e.g. 2020W12
 *
 * @return bool True if the given week started at least 9 days ago (aiming for Wednesday payouts).
 * @throws Exception
 */
function bt_week_can_be_paid( $year_week ) {
  // make a DateTime object from the given year/week, which makes it be 0:00 on Monday of that week
  $given_week = new \DateTime( $year_week, new \DateTimeZone( get_option( 'timezone_string' ) ) );

  // calculate how many days ago the start of that week was
  $days_ago = $given_week->diff( new \DateTime() )->days;

  //error_log( __METHOD__ . ": interval = " . print_r( $given_week->diff( new \DateTime() ), 1 ) );
  //error_log( __METHOD__ . ": diff days = $days_ago" );

  // we aim for 9 days later, doing payouts on the Wednesday of the following week
  return $days_ago >= 9;
}
